xquery version "1.0-ml";

import module namespace admin = "http://marklogic.com/xdmp/admin" 
   at "/MarkLogic/admin.xqy";

declare namespace xdmp = "http://marklogic.com/xdmp";
declare namespace mt   = "http://marklogic.com/xdmp/mimetypes";

declare function local:type-from-filename($name as xs:string) as xs:string*
{
   let $ext   := fn:tokenize($name, '\.')[fn:last()]
   let $types := admin:mimetypes-get(admin:get-configuration())
   return
      $types[mt:extensions/data() = $ext]/mt:name
};

(:~
 : Parse a HTTP Content-Type value.
 :
 : $ctype must conform to RFC 2616 grammar for Content-Type.  The returned
 : element looks like the following:
 :
 : <!-- result of parsing "text/plain;charset=windows-1250" -->
 : <content-type type="text" subtype="plain">
 :    <param name="charset" value="windows-1250"/>
 : </content-type>
 : 
 : # https://tools.ietf.org/html/rfc2616#section-3.7
 :
 : media-type     = type "/" subtype *( ";" parameter )
 : type           = token
 : subtype        = token
 :
 : The rule `parameter` is defined as `attribute "=" value`.  This function
 : allows space characters between ";", `attribute`, "=", and `value`, and
 : simply ignore them.
 :)
declare function local:parse-content-type($ctype as xs:string)
   as element(content-type)
{
   if ( fn:contains($ctype, '/') ) then
      <content-type type="{ fn:substring-before($ctype, '/') }"> {
         local:parse-content-type-1(
            fn:substring-after($ctype, '/'))
      }
      </content-type>
   else
      local:error('content-type-no-slash', 'invalid content-type, no slash: ' || $ctype)
};

(:~
 : Private helper for `local:parse-content-type()`.
 :)
declare function local:parse-content-type-1($input as xs:string)
   as node()+
{
   if ( fn:contains($input, ';') ) then (
      attribute { 'subtype' } { fn:substring-before($input, ';') },
      local:parse-content-type-2(
         fn:string-to-codepoints(
            fn:substring-after($input, ';')),
         (), (), 0)
   )
   else (
      attribute { 'subtype' } { $input }
   )
};

(:~
 : Private helper for `local:parse-content-type-1()`.
 :
 : # https://tools.ietf.org/html/rfc2616#section-3.6
 :
 : parameter               = attribute "=" value
 : attribute               = token
 : value                   = token | quoted-string
 :
 : # https://tools.ietf.org/html/rfc2616#section-2.2
 :
 : token          = 1*<any CHAR except CTLs or separators>
 : separators     = "(" | ")" | "<" | ">" | "@"
 :                | "," | ";" | ":" | "\" | <">
 :                | "/" | "[" | "]" | "?" | "="
 :                | "{" | "}" | SP | HT
 :
 : quoted-string  = ( <"> *(qdtext | quoted-pair ) <"> )
 : qdtext         = <any TEXT except <">>
 : quoted-pair    = "\" CHAR
 :
 : # Some char codes
 :
 : space=32
 : double quote=34
 : semi colon=59
 : equals=61
 : backslash=92
 :
 : # States (values for $state)
 :
 : 0: initial state, after a ';' has been seen, looking for a name
 : 1: scanning a name
 : 2: scanning a token
 : 3: scanning the content of a quoted string
 : 4: after the closing '"' of a quoted string
 : 5: after a '=' has been seen, looking for a value
 : 6: scanning spaces (before and after name, before and after value)
 :)
declare function local:parse-content-type-2(
   $input as xs:integer*,
   $name  as xs:integer*,
   $value as xs:integer*,
   $state as xs:integer
) as element(param)+
{
   let $head := fn:head($input)
   let $tail := fn:tail($input)
   return
      if ( fn:empty($input) ) then (
         local:parse-content-type-3($name, $value, $state)
      )
      else if ( $head eq 32 and $state = (0, 5, 6) ) then (
         local:parse-content-type-2($tail, $name, $value, $state)
      )
      else if ( $head eq 32 and $state = (1, 2, 4) ) then (
         local:parse-content-type-2($tail, $name, $value, 6)
      )
      else if ( $state eq 6 ) then (
         local:error('content-type-invalid-char', 'invalid char whilst consuming spaces: ' || $head)
      )
      else if ( $head eq 61 and $state ne 3 ) then (
         if ( fn:empty($name) ) then
            local:error('content-type-empty-name', 'empty name when encountering equals: ' || $state)
         else
            local:parse-content-type-2($tail, $name, $value, 5)
      )
      else if ( $head eq 59 and $state ne 3 ) then (
            local:parse-content-type-3($name, $value, $state),
            local:parse-content-type-2($tail, (), (), 1)
      )
      else if ( $state eq 4 ) then (
         local:error('content-type-invalid-char', 'invalid char after quoted string ended: ' || $head)
      )
      else if ( $state = (0, 1) ) then (
         local:parse-content-type-2($tail, ($name, $head), $value, 1)
      )
      else if ( $state eq 5 and $head eq 34 ) then (
         local:parse-content-type-2($tail, $name, $value, 3)
      )
      else if ( $state eq 5 ) then (
         local:parse-content-type-2($tail, $name, ($value, $head), 2)
      )
      else if ( $state eq 3 and $head eq 92 ) then (
         local:parse-content-type-2(fn:tail($tail), $name, ($value, $input[2]), $state)
      )
      else if ( $state eq 3 and $head eq 34 ) then (
         local:parse-content-type-2($tail, $name, $value, 4)
      )
      else if ( $head eq 34 ) then (
         local:error('content-type-invalid-char', 'double quote in invalid state: ' || $state)
      )
      else (
         local:parse-content-type-2($tail, $name, ($value, $head), $state)
      )
};

(:~
 : Private helper for `local:parse-content-type-2()`.
 :)
declare function local:parse-content-type-3(
   $name  as xs:integer*,
   $value as xs:integer*,
   $state as xs:integer
) as element(param)
{
   if ( $state = (2, 4) ) then
      <param name="{ fn:codepoints-to-string($name) }" value="{ fn:codepoints-to-string($value) }"/>
   else
      local:error('content-type-invalid-state', 'invalid state when semi-colon or <eof>: ' || $state)
};

(:~
 : Helper function to throw an error with a code in local:*.
 :
 : Indeed, should use another namespace than local:*...
 :)
declare function local:error($code as xs:string, $msg as xs:string)
{
   fn:error(xs:QName('local:' || $code), $msg)
};

<fields version="{ xdmp:version() }">
{
   for $name     in xdmp:get-request-field-names()
   let $value    := xdmp:get-request-field($name)
   let $filename := xdmp:get-request-field-filename($name)
   let $ctype    := xdmp:get-request-field-content-type($name)
   return
      <field>
         <name>{ $name }</name>
         {
            if ( fn:empty($value) ) then (
               <is-empty>true</is-empty>
            )
            else if ( $value instance of node()+ ) then (
               <exists-elem>{ fn:exists($value/*) }</exists-elem>,
               <exists-text>{ fn:exists($value/text()) }</exists-text>,
               <empty-2nd>{ fn:empty($value/node()[2]) }</empty-2nd>,
               <is-text>{ $value/node() instance of text() }</is-text>,
               <is-binary>{ $value/node() instance of binary() }</is-binary>
            )
            else (
               <is-atomic>true</is-atomic>
            )
         }
         <filename type="{ local:type-from-filename($filename) }">{ $filename }</filename>
         {
            local:parse-content-type($ctype)
         }
         {
            if ( $value instance of binary() ) then
               <value>...binary...</value>
            else
               <value>{ $value }</value>
         }
      </field>
}
</fields>
